"""Copyright 2019 Cisco Systems
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

The contents of this file are licensed under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""

"""Builder to ease constructing cisco_gnmi.Client derived classes."""

import logging

import grpc
from . import Client, XRClient, NXClient, XEClient
from .auth import CiscoAuthPlugin
from .util import gen_target_netloc, get_cert_from_target, get_cn_from_cert


class ClientBuilder(object):
    """Builder for the creation of a gNMI client.
    Supports construction of base Client and XRClient.
    Returns itself after each build stage to support chained initialization.

    Methods
    -------
    set_target(...)
        Specifies the network element to build a client for.
    set_os(...)
        Specifies which OS wrapper to deliver.
    set_secure(...)
        Specifies that a secure gRPC channel should be used.
    set_secure_from_file(...)
        Loads certificates from file system for secure gRPC channel.
    set_secure_from_target(...)
        Attempts to utilize available certificate from target for secure gRPC channel.
    set_call_authentication(...)
        Specifies username/password to utilize for authentication.
    set_ssl_target_override(...)
        Sets the gRPC option to override the SSL target name.
    set_channel_option(...)
        Sets a gRPC channel option. Implies knowledge of channel options.
    construct()
        Constructs and returns the built Client.
    _reset()
        Resets builder to baseline state.
    
    Examples
    --------
    >>> from cisco_gnmi import ClientBuilder
    >>> client = ClientBuilder(
    ...     '127.0.0.1:9339'
    ... ).set_os(
    ...     'IOS XR'
    ... ).set_secure_from_target().set_ssl_target_override().set_authentication(
    ...     'admin',
    ...     'its_a_secret'
    ... ).construct()
    >>> capabilities = client.capabilities()
    >>> print(capabilities)
    """

    os_class_map = {
        None: Client,
        "IOS XR": XRClient,
        "NX-OS": NXClient,
        "IOS XE": XEClient,
    }

    def __init__(self, target):
        """Initializes the builder, most initialization is done via set_* methods.
        A target is always required, thus a member of the constructor.

        Parameters
        ----------
        target : str
            The target address of the network element to interact with.
            Expects a URL-like form, e.g. 127.0.0.1:9339
        """
        self.set_target(target)
        self._reset()

    def set_target(self, target):
        """Specifies the network element to build a client for.

        Parameters
        ----------
        target : str
            The target address of the network element to interact with.
            Expects a URL-like form, e.g. 127.0.0.1:9339
        
        Returns
        -------
        self
        """
        self.__target = target
        self.__target_netloc = gen_target_netloc(self.__target)
        return self

    def set_os(self, name=None):
        """Sets which OS to target which maps to an OS wrapper class.

        Parameters
        ----------
        name : str
            "IOS XR" maps to the XRClient class.
            "NX-OS" maps to the NXClient class.
            "IOS XE" maps to the XEClient class.
            None maps to the base Client class which simply wraps the gNMI stub.
            ["IOS XR", "NX-OS", "IOS XE", None]
        
        Returns
        -------
        self
        """
        if name not in self.os_class_map.keys():
            raise Exception("OS not supported!")
        else:
            self.__client_class = self.os_class_map[name]
        logging.debug("Using %s wrapper.", name or "Client")
        return self

    def set_secure(
        self, root_certificates=None, private_key=None, certificate_chain=None
    ):
        """Simply sets the fields to be expected by grpc.ssl_channel_credentials(...).
        Setting this method disallows following set_secure(...) or _set_insecure() calls.

        Parameters
        ----------
        root_certificates : str or None
        private_key : str or None
        certificate_chain : str or None

        Returns
        -------
        self
        """
        self.__root_certificates = root_certificates
        self.__private_key = private_key
        self.__certificate_chain = certificate_chain
        return self

    def set_secure_from_file(
        self, root_certificates=None, private_key=None, certificate_chain=None
    ):
        """Wraps set_secure(...) but treats arguments as file paths.

        Parameters
        ----------
        root_certicates : str or None
        private_key : str or None
        certificate_chain : str or None
        
        Returns
        -------
        self
        """

        def load_cert(file_path):
            cert_content = None
            if file_path is not None:
                with open(file_path, "rb") as cert_fd:
                    cert_content = cert_fd.read()
            return cert_content

        if root_certificates:
            root_certificates = load_cert(root_certificates)
        if private_key:
            private_key = load_cert(private_key)
        if certificate_chain:
            certificate_chain = load_cert(certificate_chain)
        self.set_secure(root_certificates, private_key, certificate_chain)
        return self

    def set_secure_from_target(self):
        """Wraps set_secure(...) but loads root certificates from target.
        In effect, simply uses the target's certificate to create an encrypted channel.

        TODO: This may not work with IOS XE and NX-OS, uncertain.

        Returns
        -------
        self
        """
        root_certificates = get_cert_from_target(self.__target_netloc)
        self.set_secure(root_certificates)
        return self

    def set_call_authentication(self, username, password):
        """Sets the username and password to utilize for authentication."""
        self.__username = username
        self.__password = password
        return self

    def set_ssl_target_override(self, ssl_target_name_override=None):
        """Sets the gRPC option to override the SSL target name with the specified value.
        
        If None supplied then the option will be derived from the CN of the root certificate.
        set_secure_from_target().set_ssl_target_override() effectively creates an encrypted
        but insecure channel.
        The above behavior attempts to replicate the Go gNMI reference -insecure parameter.
        This is not recommended other than for testing purposes.

        Parameters
        ----------
        ssl_target_name_override : str or None
            Value for grpc.ssl_target_name_override
        
        Returns
        -------
        self
        """
        self.__ssl_target_name_override = ssl_target_name_override
        return self

    def set_channel_option(self, name, value):
        """Sets a gRPC channel option. This method implies understanding of gRPC channels.
        If the option is found, the value is overwritten.

        Parameters
        ----------
        name : str
            The gRPC channel option name.
        value : ?
            The value of the named option.

        Returns
        -------
        self
        """
        new_option = (name, value)
        if not self.__channel_options:
            self.__channel_options = [new_option]
        else:
            found_index = None
            for index, option in enumerate(self.__channel_options):
                if option[0] == name:
                    found_index = index
                    break
            if found_index is not None:
                logging.warning("Found existing channel option %s, overwriting!", name)
                self.__channel_options[found_index] = new_option
            else:
                self.__channel_options.append(new_option)
        return self

    def construct(self):
        """Constructs and returns the desired Client object.
        The instance of this class will reset to default values for further building.

        Returns
        -------
        Client or NXClient or XEClient or XRClient
        """
        channel = None
        channel_ssl_creds = None
        channel_metadata_creds = None
        channel_creds = None
        channel_ssl_creds = grpc.ssl_channel_credentials(
            self.__root_certificates, self.__private_key, self.__certificate_chain
        )
        if self.__username and self.__password:
            channel_metadata_creds = grpc.metadata_call_credentials(
                CiscoAuthPlugin(self.__username, self.__password)
            )
            logging.debug("Using username/password call authentication.")
        if channel_ssl_creds and channel_metadata_creds:
            channel_creds = grpc.composite_channel_credentials(
                channel_ssl_creds, channel_metadata_creds
            )
            logging.debug("Using SSL/metadata authentication composite credentials.")
        else:
            channel_creds = channel_ssl_creds
            logging.debug("Using SSL credentials, no metadata authentication.")
        if self.__ssl_target_name_override is not False:
            if self.__ssl_target_name_override is None:
                if not self.__root_certificates:
                    raise Exception("Deriving override requires root certificate!")
                self.__ssl_target_name_override = get_cn_from_cert(
                    self.__root_certificates
                )
                logging.warning(
                    "Overriding SSL option from certificate could increase MITM susceptibility!"
                )
            self.set_channel_option(
                "grpc.ssl_target_name_override", self.__ssl_target_name_override
            )
        channel = grpc.secure_channel(
            self.__target_netloc.netloc, channel_creds, self.__channel_options
        )
        if self.__client_class is None:
            self.set_os()
        client = self.__client_class(channel)
        self._reset()
        return client

    def _reset(self):
        """Resets the builder.
        
        Returns
        -------
        self
        """
        self.set_target(self.__target)
        self.__client_class = None
        self.__root_certificates = None
        self.__private_key = None
        self.__certificate_chain = None
        self.__username = None
        self.__password = None
        self.__channel_options = None
        self.__ssl_target_name_override = False
        return self



"""Copyright 2019 Cisco Systems
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

The contents of this file are licensed under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""

"""Wrapper for NX-OS to simplify usage of gNMI implementation."""


import logging

from six import string_types
from .client import Client, proto, util


class NXClient(Client):
    """NX-OS-specific wrapper for gNMI functionality.

    Returns direct responses from base Client methods.

    Methods
    -------
    subscribe_xpaths(...)
        Convenience wrapper for subscribe() which helps construct subscriptions for specified xpaths.

    Examples
    --------
    >>> from cisco_gnmi import ClientBuilder
    >>> client = ClientBuilder('127.0.0.1:9339').set_os(
    ...     'NX-OS'
    ... ).set_secure_from_file().set_ssl_target_override().set_call_authentication(
    ...     'admin',
    ...     'its_a_secret'
    ... ).construct()
    >>> capabilities = client.capabilities()
    >>> print(capabilities)
    """

    def get(self, *args, **kwargs):
        raise NotImplementedError("Get not yet supported on NX-OS!")

    def set(self, *args, **kwargs):
        raise NotImplementedError("Set not yet supported on NX-OS!")

    def subscribe_xpaths(
        self,
        xpath_subscriptions,
        request_mode="STREAM",
        sub_mode="SAMPLE",
        encoding="PROTO",
        sample_interval=Client._NS_IN_S * 10,
        suppress_redundant=False,
        heartbeat_interval=None,
    ):
        """A convenience wrapper of subscribe() which aids in building of SubscriptionRequest
        with request as subscribe SubscriptionList. This method accepts an iterable of simply xpath strings,
        dictionaries with Subscription attributes for more granularity, or already built Subscription
        objects and builds the SubscriptionList. Fields not supplied will be defaulted with the default arguments
        to the method.

        Generates a single SubscribeRequest.

        Parameters
        ----------
        xpath_subscriptions : str or iterable of str, dict, Subscription
            An iterable which is parsed to form the Subscriptions in the SubscriptionList to be passed
            to SubscriptionRequest. Strings are parsed as XPaths and defaulted with the default arguments,
            dictionaries are treated as dicts of args to pass to the Subscribe init, and Subscription is
            treated as simply a pre-made Subscription.
        request_mode : proto.gnmi_pb2.SubscriptionList.Mode, optional
            Indicates whether STREAM to stream from target,
            ONCE to stream once (like a get),
            POLL to respond to POLL.
            [STREAM, ONCE, POLL]
        sub_mode : proto.gnmi_pb2.SubscriptionMode, optional
            The default SubscriptionMode on a per Subscription basis in the SubscriptionList.
            ON_CHANGE only streams updates when changes occur.
            SAMPLE will stream the subscription at a regular cadence/interval.
            [ON_CHANGE, SAMPLE]
        encoding : proto.gnmi_pb2.Encoding, optional
            A member of the proto.gnmi_pb2.Encoding enum specifying desired encoding of returned data
            [JSON, PROTO]
        sample_interval : int, optional
            Default nanoseconds for sample to occur.
            Defaults to 10 seconds.
        suppress_redundant : bool, optional
            Indicates whether values that have not changed should be sent in a SAMPLE subscription.
        heartbeat_interval : int, optional
            Specifies the maximum allowable silent period in nanoseconds when
            suppress_redundant is in use. The target should send a value at least once
            in the period specified.

        Returns
        -------
        subscribe()
        """
        supported_request_modes = ["STREAM", "ONCE", "POLL"]
        request_mode = util.validate_proto_enum(
            "mode",
            request_mode,
            "SubscriptionList.Mode",
            proto.gnmi_pb2.SubscriptionList.Mode,
            subset=supported_request_modes,
            return_name=True,
        )
        supported_encodings = ["JSON", "PROTO"]
        encoding = util.validate_proto_enum(
            "encoding",
            encoding,
            "Encoding",
            proto.gnmi_pb2.Encoding,
            subset=supported_encodings,
            return_name=True,
        )
        supported_sub_modes = ["ON_CHANGE", "SAMPLE"]
        sub_mode = util.validate_proto_enum(
            "sub_mode",
            sub_mode,
            "SubscriptionMode",
            proto.gnmi_pb2.SubscriptionMode,
            subset=supported_sub_modes,
            return_name=True,
        )
        return super(NXClient, self).subscribe_xpaths(
            xpath_subscriptions,
            request_mode,
            sub_mode,
            encoding,
            sample_interval,
            suppress_redundant,
            heartbeat_interval,
        )

    def parse_xpath_to_gnmi_path(self, xpath, origin=None):
        """Attempts to determine whether origin should be YANG (device) or DME.
        Errors on OpenConfig until support is present.
        """
        if xpath.startswith("openconfig"):
            raise NotImplementedError(
                "OpenConfig data models not yet supported on NX-OS!"
            )
        if origin is None:
            if any(
                map(xpath.startswith, ["Cisco-NX-OS-device", "/Cisco-NX-OS-device"])
            ):
                origin = "device"
                # Remove the module
                xpath = xpath.split(":", 1)[1]
            else:
                origin = "DME"
        return super(NXClient, self).parse_xpath_to_gnmi_path(xpath, origin)



"""Copyright 2019 Cisco Systems
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

The contents of this file are licensed under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""

"""Wrapper for IOS XR to simplify usage of gNMI implementation."""

import json
import logging

from six import string_types
from .client import Client, proto, util


class XRClient(Client):
    """IOS XR-specific wrapper for gNMI functionality.

    Returns direct responses from base Client methods.

    Methods
    -------
    delete_xpaths(...)
        Convenience wrapper for set() which constructs Paths from XPaths for deletion.
    get_xpaths(...)
        Convenience wrapper for get() which helps construct get requests for specified xpaths.
    set_json(...)
        Convenience wrapper for set() which assumes model-based JSON payloads.
    subscribe_xpaths(...)
        Convenience wrapper for subscribe() which helps construct subscriptions for specified xpaths.

    Examples
    --------
    >>> from cisco_gnmi import ClientBuilder
    >>> client = ClientBuilder('127.0.0.1:9339').set_os(
    ...     'IOS XR'
    ... ).set_secure_from_file(
    ...     'ems.pem',
    ... ).set_ssl_target_override().set_call_authentication(
    ...     'admin',
    ...     'its_a_secret'
    ... ).construct()
    >>> capabilities = client.capabilities()
    >>> print(capabilities)
    ...
    >>> get_response = client.get_xpaths('interfaces/interface')
    >>> print(get_response)
    ...
    >>> subscribe_response = client.subscribe_xpaths('interfaces/interface')
    >>> for message in subscribe_response: print(message)
    ...
    >>> config = '{"Cisco-IOS-XR-shellutil-cfg:host-names": [{"host-name": "gnmi_test"}]}'
    >>> set_response = client.set_json(config)
    >>> print(set_response)
    ...
    >>> delete_response = client.delete_xpaths('Cisco-IOS-XR-shellutil-cfg:host-names/host-name')
    """

    def delete_xpaths(self, xpaths, prefix=None):
        """A convenience wrapper for set() which constructs Paths from supplied xpaths
        to be passed to set() as the delete parameter.

        Parameters
        ----------
        xpaths : iterable of str
            XPaths to specify to be deleted.
            If prefix is specified these strings are assumed to be the suffixes.
        prefix : str
            The XPath prefix to apply to all XPaths for deletion.

        Returns
        -------
        set()
        """
        if isinstance(xpaths, string_types):
            xpaths = [xpaths]
        paths = []
        for xpath in xpaths:
            if prefix:
                if prefix.endswith("/") and xpath.startswith("/"):
                    xpath = "{prefix}{xpath}".format(
                        prefix=prefix[:-1], xpath=xpath[1:]
                    )
                elif prefix.endswith("/") or xpath.startswith("/"):
                    xpath = "{prefix}{xpath}".format(prefix=prefix, xpath=xpath)
                else:
                    xpath = "{prefix}/{xpath}".format(prefix=prefix, xpath=xpath)
            paths.append(self.parse_xpath_to_gnmi_path(xpath))
        return self.set(deletes=paths)

    def set_json(self, update_json_configs=None, replace_json_configs=None, ietf=True):
        """A convenience wrapper for set() which assumes JSON payloads and constructs desired messages.
        All parameters are optional, but at least one must be present.

        This method expects JSON in the same format as what you might send via the native gRPC interface
        with a fully modeled configuration which is then parsed to meet the gNMI implementation.

        Parameters
        ----------
        update_json_configs : iterable of JSON configurations, optional
            JSON configs to apply as updates.
        replace_json_configs : iterable of JSON configurations, optional
            JSON configs to apply as replacements.
        ietf : bool, optional
            Use JSON_IETF vs JSON.

        Returns
        -------
        set()
        """
        if not any([update_json_configs, replace_json_configs]):
            raise Exception("Must supply at least one set of configurations to method!")

        def check_configs(name, configs):
            if isinstance(name, string_types):
                logging.debug("Handling %s as JSON string.", name)
                try:
                    configs = json.loads(configs)
                except:
                    raise Exception("{name} is invalid JSON!".format(name=name))
                configs = [configs]
            elif isinstance(name, dict):
                logging.debug("Handling %s as already serialized JSON object.", name)
                configs = [configs]
            elif not isinstance(configs, (list, set)):
                raise Exception(
                    "{name} must be an iterable of configs!".format(name=name)
                )
            return configs

        def create_updates(name, configs):
            if not configs:
                return None
            configs = check_configs(name, configs)
            updates = []
            for config in configs:
                if not isinstance(config, dict):
                    raise Exception("config must be a JSON object!")
                if len(config.keys()) > 1:
                    raise Exception("config should only target one YANG module!")
                top_element = next(iter(config.keys()))
                top_element_split = top_element.split(":")
                if len(top_element_split) < 2:
                    raise Exception(
                        "Top level config element {} should be module prefixed!".format(
                            top_element
                        )
                    )
                if len(top_element_split) > 2:
                    raise Exception(
                        "Top level config element {} appears malformed!".format(
                            top_element
                        )
                    )
                origin = top_element_split[0]
                element = top_element_split[1]
                config = config.pop(top_element)
                update = proto.gnmi_pb2.Update()
                update.path.CopyFrom(self.parse_xpath_to_gnmi_path(element, origin))
                if ietf:
                    update.val.json_ietf_val = json.dumps(config).encode("utf-8")
                else:
                    update.val.json_val = json.dumps(config).encode("utf-8")
                updates.append(update)
            return updates

        updates = create_updates("update_json_configs", update_json_configs)
        replaces = create_updates("replace_json_configs", replace_json_configs)
        return self.set(updates=updates, replaces=replaces)

    def get_xpaths(self, xpaths, data_type="ALL", encoding="JSON_IETF"):
        """A convenience wrapper for get() which forms proto.gnmi_pb2.Path from supplied xpaths.

        Parameters
        ----------
        xpaths : iterable of str or str
            An iterable of XPath strings to request data of
            If simply a str, wraps as a list for convenience
        data_type : proto.gnmi_pb2.GetRequest.DataType, optional
            A direct value or key from the GetRequest.DataType enum
            [ALL, CONFIG, STATE, OPERATIONAL]
        encoding : proto.gnmi_pb2.GetRequest.Encoding, optional
            A direct value or key from the Encoding enum
            [JSON, BYTES, PROTO, ASCII, JSON_IETF]

        Returns
        -------
        get()
        """
        gnmi_path = None
        if isinstance(xpaths, (list, set)):
            gnmi_path = map(self.parse_xpath_to_gnmi_path, set(xpaths))
        elif isinstance(xpaths, string_types):
            gnmi_path = [self.parse_xpath_to_gnmi_path(xpaths)]
        else:
            raise Exception(
                "xpaths must be a single xpath string or iterable of xpath strings!"
            )
        return self.get(gnmi_path, data_type=data_type, encoding=encoding)

    def get_cli(self, commands):
        """A convenience wrapper for get() which forms proto.gnmi_pb2.Path from supplied CLI commands.
        IOS XR appears to be the only OS with this functionality.

        Parameters
        ----------
        commands : iterable of str or str
            An iterable of CLI commands as strings to request data of
            If simply a str, wraps as a list for convenience

        Returns
        -------
        get()
        """
        gnmi_path = None
        if isinstance(commands, (list, set)):
            gnmi_path = list(map(self.parse_cli_to_gnmi_path, commands))
        elif isinstance(commands, string_types):
            gnmi_path = [self.parse_cli_to_gnmi_path(commands)]
        else:
            raise Exception(
                "commands must be a single CLI command string or iterable of CLI commands as strings!"
            )
        return self.get(gnmi_path, encoding="ASCII")

    def subscribe_xpaths(
        self,
        xpath_subscriptions,
        request_mode="STREAM",
        sub_mode="SAMPLE",
        encoding="PROTO",
        sample_interval=Client._NS_IN_S * 10,
        suppress_redundant=False,
        heartbeat_interval=None,
    ):
        """A convenience wrapper of subscribe() which aids in building of SubscriptionRequest
        with request as subscribe SubscriptionList. This method accepts an iterable of simply xpath strings,
        dictionaries with Subscription attributes for more granularity, or already built Subscription
        objects and builds the SubscriptionList. Fields not supplied will be defaulted with the default arguments
        to the method.

        Generates a single SubscribeRequest.

        Parameters
        ----------
        xpath_subscriptions : str or iterable of str, dict, Subscription
            An iterable which is parsed to form the Subscriptions in the SubscriptionList to be passed
            to SubscriptionRequest. Strings are parsed as XPaths and defaulted with the default arguments,
            dictionaries are treated as dicts of args to pass to the Subscribe init, and Subscription is
            treated as simply a pre-made Subscription.
        request_mode : proto.gnmi_pb2.SubscriptionList.Mode, optional
            Indicates whether STREAM to stream from target,
            ONCE to stream once (like a get),
            POLL to respond to POLL.
            [STREAM, ONCE, POLL]
        sub_mode : proto.gnmi_pb2.SubscriptionMode, optional
            The default SubscriptionMode on a per Subscription basis in the SubscriptionList.
            TARGET_DEFINED indicates that the target (like device/destination) should stream
            information however it knows best. This instructs the target to decide between ON_CHANGE
            or SAMPLE - e.g. the device gNMI server may understand that we only need RIB updates
            as an ON_CHANGE basis as opposed to SAMPLE, and we don't have to explicitly state our
            desired behavior.
            ON_CHANGE only streams updates when changes occur.
            SAMPLE will stream the subscription at a regular cadence/interval.
            [ON_CHANGE, SAMPLE]
        encoding : proto.gnmi_pb2.Encoding, optional
            A member of the proto.gnmi_pb2.Encoding enum specifying desired encoding of returned data
            [PROTO]
        sample_interval : int, optional
            Default nanoseconds for sample to occur.
            Defaults to 10 seconds.
        suppress_redundant : bool, optional
            Indicates whether values that have not changed should be sent in a SAMPLE subscription.
        heartbeat_interval : int, optional
            Specifies the maximum allowable silent period in nanoseconds when
            suppress_redundant is in use. The target should send a value at least once
            in the period specified.

        Returns
        -------
        subscribe()
        """
        supported_request_modes = ["STREAM", "ONCE", "POLL"]
        request_mode = util.validate_proto_enum(
            "mode",
            request_mode,
            "SubscriptionList.Mode",
            proto.gnmi_pb2.SubscriptionList.Mode,
            subset=supported_request_modes,
            return_name=True,
        )
        supported_encodings = ["PROTO"]
        encoding = util.validate_proto_enum(
            "encoding",
            encoding,
            "Encoding",
            proto.gnmi_pb2.Encoding,
            subset=supported_encodings,
            return_name=True,
        )
        supported_sub_modes = ["ON_CHANGE", "SAMPLE"]
        sub_mode = util.validate_proto_enum(
            "sub_mode",
            sub_mode,
            "SubscriptionMode",
            proto.gnmi_pb2.SubscriptionMode,
            subset=supported_sub_modes,
            return_name=True,
        )
        return super(XRClient, self).subscribe_xpaths(
            xpath_subscriptions,
            request_mode,
            sub_mode,
            encoding,
            sample_interval,
            suppress_redundant,
            heartbeat_interval,
        )

    def parse_xpath_to_gnmi_path(self, xpath, origin=None):
        """No origin specified implies openconfig
        Otherwise origin is expected to be the module name
        """
        if origin is None:
            # naive but effective
            if xpath.startswith("openconfig") or ":" not in xpath:
                # openconfig
                origin = None
            else:
                # module name
                origin, xpath = xpath.split(":", 1)
                origin = origin.strip("/")
        return super(XRClient, self).parse_xpath_to_gnmi_path(xpath, origin)

    def parse_cli_to_gnmi_path(self, command):
        """Parses a CLI command to proto.gnmi_pb2.Path.
        IOS XR appears to be the only OS with this functionality.

        The CLI command becomes a path element.
        """
        if not isinstance(command, string_types):
            raise Exception("command must be a string!")
        path = proto.gnmi_pb2.Path()
        curr_elem = proto.gnmi_pb2.PathElem()
        curr_elem.name = command
        path.elem.extend([curr_elem])
        return path



#!/usr/bin/env python
"""Copyright 2020 Cisco Systems
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

The contents of this file are licensed under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""

"""Derived from Flask
https://github.com/pallets/flask/blob/master/setup.py
"""

import io
import re

from setuptools import find_packages
from setuptools import setup

with io.open("README.md", "rt", encoding="utf8") as f:
    readme = f.read()

with io.open("src/cisco_gnmi/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setup(
    name="cisco_gnmi",
    version=version,
    url="https://github.com/cisco-ie/cisco-gnmi-python",
    project_urls={
        "Code": "https://github.com/cisco-ie/cisco-gnmi-python",
        "Issue Tracker": "https://github.com/cisco-ie/cisco-gnmi-python/issues",
    },
    license="Apache License (2.0)",
    author="Cisco Innovation Edge",
    author_email="cisco-ie@cisco.com",
    maintainer="Cisco Innovation Edge",
    maintainer_email="cisco-ie@cisco.com",
    description="This library wraps gNMI functionality to ease usage with Cisco implementations.",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: System :: Networking",
        "Topic :: System :: Networking :: Monitoring",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4",
    install_requires=[
        "grpcio",
        "protobuf",
        "six",
        "cryptography",
    ],
    extras_require={
        "dev": [
            "grpcio-tools",
            "googleapis-common-protos",
            "pylint",
            "twine",
            "setuptools",
            "wheel",
            "pytest",
            "pytest-cov",
            "pytest-mock",
            "coverage",
        ],
    },
    entry_points={"console_scripts": ["cisco-gnmi = cisco_gnmi.cli:main"]},
)
